package org.netkernel.layer1.endpoint;

import org.netkernel.container.IKernel;
import org.netkernel.container.ISpaceListener;
import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.util.GoldenThreadExpiryFunction;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.impl.RequestScopeLevelImpl;
import org.netkernel.urii.ISpace;
import org.netkernel.urii.ISpaceWithIdentity;

import java.util.*;

public class SpaceHDSAggregator extends StandardAccessorImpl implements ISpaceListener
{
	private GoldenThreadExpiryFunction mExpiry;
	
	public SpaceHDSAggregator()
	{	this.declareThreadSafe();
		this.declareSourceRepresentation(IHDSNode.class);
		mExpiry=new GoldenThreadExpiryFunction("ISpaceListener");
	}
	
	public void onCommissionEndpoint(IKernel aKernel, ISpace aSpace) throws Exception
	{	super.onCommissionEndpoint(aKernel, aSpace);
		aKernel.addSpaceListener(this);
	}
	
	public void onDecommissionEndpoint() throws Exception
	{	getKernel().removeSpaceListener(this);
		super.onDecommissionEndpoint();
	}
	
	public void spaceChanged()
	{	mExpiry.invalidate();
		mExpiry=new GoldenThreadExpiryFunction("ISpaceListener");
	}
	
	@Override
	public void onSource(INKFRequestContext context) throws Exception
	{	
		Set<ISpace> superStack=null;
		if (context.getThisRequest().argumentExists("exclusions"))
		{	superStack=(Set)context.source("arg:exclusions",Set.class);
		}
		
		IKernel kernel=context.getKernelContext().getKernel();
		HDSBuilder b=new HDSBuilder();
		b.pushNode("spaces");
		String uri=context.getThisRequest().getArgumentValue("uri");
		b.addNode("uri", uri);
		Iterator<ISpaceWithIdentity> it=kernel.getSpaces();
		boolean errors=false;
		boolean expired=false;
		while(it.hasNext())
		{	ISpaceWithIdentity space=it.next();
			b.pushNode("space");
			b.addNode("id", space.getIdentifier().toString());
			b.addNode("version", space.getVersion().toString());
			
			if (superStack!=null)
			{	if (superStack.contains(space))
				{	b.popNode();
					continue;
				}
			}
			
			IRequestScopeLevel scope=RequestScopeLevelImpl.createOrphanedRootScopeLevel(space, context.getKernelContext().getRequestScope());
			INKFRequest req=context.createRequest(uri);
			req.setRequestScope(scope);
			Object rep;
			try
			{	INKFResponseReadOnly resp=context.issueRequestForResponse(req);
				expired|=resp.isExpired();
				rep=resp.getRepresentation();
			}
			catch(Exception e)
			{	//Space doesn't contain the resource
				rep=null;
			}
			if (rep!=null)
			{	try
				{	if (!(rep instanceof IHDSNode))
					{	rep=(IHDSNode)context.transrept(rep, IHDSNode.class);
					}
					b.importChildren(((IHDSNode)rep));
				}
				catch (NKFException e)
				{	context.logFormatted(INKFLocale.LEVEL_WARNING, "MSG_SA_IMP_BAD_TYPE", rep.getClass().getName(),space.getIdentifier().toString(),uri);
					errors=true;
				}
			}
			b.popNode();
		}
		b.popNode();
		INKFResponse resp=context.createResponseFrom(b.getRoot());
		if (errors)
		{	resp.setExpiry(INKFResponse.EXPIRY_FUNCTION,mExpiry);
		}
		else
		{	context.getKernelContext().declareDependency(mExpiry, Integer.MAX_VALUE);
			if (expired)
			{	long expiryTime=getKernel().getTimer().getApproximateTime()+getKernel().getConfiguration().getLong("stdmod.initialisation.hysteresis", 500L);
				resp.setExpiry(INKFResponse.EXPIRY_MAX_CONSTANT_DEPENDENT, expiryTime);
			}
		}
	}
}
